@@ -4,6 +4,7 @@
 // FUNCTIONS
 int get_cell(const float& x, const float& xmin, const float& xmax, const int& nc);
 float get_velocity(const float& x, const float* A, const float& xmin, const float& xmax, const int& nc);
+float get_velocity_dx(const float& x, const float* A, const float& xmin, const float& xmax, const int& nc);
 
 // INTEGRATION
 float integrate_closed_form(float x, float t, const float* A, const float& xmin, const float& xmax, const int& nc);
@@ -16,5 +17,6 @@ void derivative_phi_theta(float* dphi_dtheta, const float& xini, const float& tm
 // GRADIENT SPACE
 float derivative_phi_x(const float& xini, const float& tini, const float& tm, const int& cm, const float* A, const float& xmin, const float& xmax, const int& nc);
 void derivative_phi_x_theta(float* dphi_dx_dtheta, const float& xini, const float& tini, const float& tm, const int& cm, const int& d, const float* B, const float* A, const float& xmin, const float& xmax, const int& nc);
+float derivative_phi_x_x(const float& xini, const float& tini, const float& tm, const int& cm, const float* A, const float& xmin, const float& xmax, const int& nc);
 
 #endif
\ No newline at end of file
